@@ -3,8 +3,11 @@ import logging
 from enum import Enum
 from logging import getLogger
 import cz_ossfs
+import s3fs
+from google.oauth2.credentials import Credentials
 from pyarrow import fs
 import string
+import gcsfs
 
 from clickzetta.bulkload import cz_table
 from clickzetta.proto.generated import ingestion_pb2, file_format_type_pb2
@@ -65,6 +68,20 @@ class StagingConfig:
         if self.type == 'oss':
             return CZOSSFileSystem(endpoint=self.endpoint, secret=self.secret, token=self.token, key=self.id,
                                    cz_feature='bulkload')
+        elif self.type == 's3':
+            client_kwargs = {'region_name': self.endpoint}
+            return s3fs.S3FileSystem(key=self.id, secret=self.secret, token=self.token, client_kwargs=client_kwargs)
+        elif self.type == 'gcs':
+            credentials = Credentials(token=self.token)
+            return gcsfs.GCSFileSystem(project='clickzetta-connector', token=credentials)
+        elif self.type == 'tos':
+            raise NotImplementedError('TOS is not supported bulkload yet')
+            # TODO wait for volces to support s3fs. Currently, it got an error as below:
+            # TypeError: _create_client() got an unexpected keyword argument 's3'
+            # import s3fs
+            # client_kwargs = {'s3': {'addressing_style': 'virtual'}}
+            # return s3fs.S3FileSystem(key=self.id, secret=self.secret, token=self.token,
+            #                          endpoint_url='https://tos-s3-cn-guangzhou.volces.com', client_kwargs=client_kwargs)
         else:
             return fs.LocalFileSystem()
 
@@ -73,18 +90,40 @@ class BulkLoadConfig:
     def __init__(self, config: ingestion_pb2.BulkLoadStreamWriterConfig):
         self.config = config
 
-    def get_staging_config(self):
+    def get_staging_config(self, prefer_internal_endpoint=False):
         staging_path = self.config.staging_path
+        staging_config = None
         if staging_path.oss_path.path != '':
             oss_path = staging_path.oss_path
+            endpoint = oss_path.oss_endpoint
+            if prefer_internal_endpoint:
+                endpoint = oss_path.oss_internal_endpoint
             staging_config = StagingConfig(oss_path.path, oss_path.sts_ak_id, oss_path.sts_ak_secret,
-                                           oss_path.sts_token,
-                                           oss_path.oss_endpoint, 'oss')
-        else:
+                                           oss_path.sts_token, endpoint, 'oss')
+        elif staging_path.cos_path.path != '':
             cos_path = staging_path.cos_path
             staging_config = StagingConfig(cos_path.path, cos_path.sts_ak_id, cos_path.sts_ak_secret,
                                            cos_path.sts_token,
                                            cos_path.cos_region, 'cos')
+        elif staging_path.s3_path.path != '':
+            s3_path = staging_path.s3_path
+            staging_config = StagingConfig(s3_path.path, s3_path.sts_ak_id, s3_path.sts_ak_secret,
+                                           s3_path.sts_token,
+                                           s3_path.s3_region, 's3')
+        elif staging_path.gcs_path.path != '':
+            gcs_path = staging_path.gcs_path
+            staging_config = StagingConfig(gcs_path.path, gcs_path.sts_ak_id, gcs_path.sts_ak_secret,
+                                           gcs_path.sts_token,
+                                           gcs_path.gcs_region, 'gcs')
+        elif staging_path.tos_path.path != '':
+            tos_path = staging_path.tos_path
+            # replace the tos:// to https://tos-s3-cn-guangzhou.volces.com
+            tos_path.path = tos_path.path.replace('tos://', 's3://')
+            staging_config = StagingConfig(tos_path.path, tos_path.sts_ak_id, tos_path.sts_ak_secret,
+                                           tos_path.sts_token,
+                                           tos_path.tos_endpoint, 'tos')
+        else:
+            raise ValueError('Staging path is not set, please check the configuration')
         return staging_config
 
     def get_file_format(self):
@@ -137,11 +176,15 @@ class BulkLoadState(Enum):
 
 
 class BulkLoadOptions:
-    def __init__(self, operation: BulkLoadOperation, partition_specs: str, record_keys: list) -> None:
+    def __init__(self, operation: BulkLoadOperation, partition_specs: str, record_keys: list,
+                 prefer_internal_endpoint=False) -> None:
         self.operation = operation
         self.partition_specs = partition_specs
         self.record_keys = record_keys
-        self._properties = {'operation': operation, 'partition_specs': partition_specs, 'record_keys': record_keys}
+        self.prefer_internal_endpoint = prefer_internal_endpoint
+        self._properties = {'operation': operation, 'partition_specs': partition_specs, 'record_keys': record_keys,
+                            'prefer_internal_endpoint': prefer_internal_endpoint}
+
 
     def to_api_repr(self) -> dict:
         return copy.deepcopy(self._properties)
@@ -171,6 +214,9 @@ class BulkLoadMetaData:
     def get_table(self):
         return self.table
 
+    def get_prefer_internal_endpoint(self):
+        return self.info.prefer_internal_endpoint
+
     def get_operation(self):
         if self.info.operation == ingestion_pb2.BulkLoadStreamOperation.BL_APPEND:
             return BulkLoadOperation.APPEND
