@@ -1,10 +1,11 @@
 import datetime
 import decimal
-import logging
 import string
 import uuid
 
 import os
+
+import pandas as pd
 import pyarrow as pa
 import pyarrow.parquet as pq
 import pyarrow.compute as pc
@@ -38,9 +39,9 @@ class BulkLoadWriter:
         self.meta_data = meta_data
         self.config = config
         self.partition_id = partition_id
-        self.staging_config = config.get_staging_config()
+        self.staging_config = config.get_staging_config(meta_data.get_prefer_internal_endpoint())
         self._process_staging_type()
-        self.file_io = config.get_staging_config().create_file_io()
+        self.file_io = config.get_staging_config(meta_data.get_prefer_internal_endpoint()).create_file_io()
         self.file_format = config.get_file_format()
         self.max_file_records = config.get_max_rows_per_file()
         self.max_file_size = config.get_max_file_size_per_file()
@@ -71,6 +72,12 @@ class BulkLoadWriter:
             cos_config = CosConfig(Region=self.staging_config.endpoint, SecretId=self.staging_config.id,
                                    SecretKey=self.staging_config.secret, Token=self.staging_config.token)
             self.cos_client = CosS3Client(cos_config)
+        elif self.staging_config.type == 's3':
+            self.location = self.staging_config.path
+        elif self.staging_config.type == 'gcs':
+            self.location = self.staging_config.path
+        else:
+            self.location = self.staging_config.path
 
     def get_stream_id(self):
         return self.meta_data.get_stream_id()
@@ -202,43 +209,39 @@ class BulkLoadWriter:
             row_size += self._estimate_type_size(column_type)
         return row_size
 
-    def _convert_data_type(self, field_name: string, field_data: list):
+    def _convert_data_type(self, field_name: str, field_data: list):
+        """Use common utility to convert input data to arrow array with correct arrow data type"""
         schema = self.meta_data.get_table().schema
         data_type = schema[field_name]
-        if data_type.category == data_type_pb2.DataTypeCategory.INT8:
-            return pc.cast(field_data, pa.int8())
-        elif data_type.category == data_type_pb2.DataTypeCategory.INT16:
-            return pc.cast(field_data, pa.int16())
-        elif data_type.category == data_type_pb2.DataTypeCategory.INT32:
-            return pc.cast(field_data, pa.int32())
-        elif data_type.category == data_type_pb2.DataTypeCategory.INT64:
-            return pc.cast(field_data, pa.int64())
-        elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT32:
-            return pc.cast(field_data, pa.float32())
-        elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT64:
-            return pc.cast(field_data, pa.float64())
-        elif data_type.category == data_type_pb2.DataTypeCategory.DECIMAL:
-            precision = data_type.decimalTypeInfo.precision
-            scale = data_type.decimalTypeInfo.scale
-            return pc.cast(field_data, pa.decimal128(precision, scale))
-        elif data_type.category == data_type_pb2.DataTypeCategory.BOOLEAN:
-            return pc.cast(field_data, pa.bool_())
-        elif data_type.category == data_type_pb2.DataTypeCategory.CHAR or \
-                data_type.category == data_type_pb2.DataTypeCategory.VARCHAR or \
-                data_type.category == data_type_pb2.DataTypeCategory.STRING:
-            return pc.cast(field_data, pa.string())
-        elif data_type.category == data_type_pb2.DataTypeCategory.DATE:
-            return pc.cast(field_data, pa.date32())
-        elif data_type.category == data_type_pb2.DataTypeCategory.TIMESTAMP_LTZ:
-            timestamp_unit = data_type.timestamp_info.tsUnit
-            if timestamp_unit == data_type_pb2.TimestampUnit.SECONDS:
-                return pc.cast(field_data, pa.timestamp('s', tz='UTC'))
-            elif timestamp_unit == data_type_pb2.TimestampUnit.MILLISECONDS:
-                return pc.cast(field_data, pa.timestamp('ms', tz='UTC'))
-            elif timestamp_unit == data_type_pb2.TimestampUnit.MICROSECONDS:
-                return pc.cast(field_data, pa.timestamp('us', tz='UTC'))
-            elif timestamp_unit == data_type_pb2.TimestampUnit.NANOSECONDS:
-                return pc.cast(field_data, pa.timestamp('ns', tz='UTC'))
+        arrow_type = to_arrow_type(data_type)
+
+        if not field_data:
+            return pa.array([], type=arrow_type)
+
+        # Convert None, NaT, NaN to None
+        batch_data = []
+        for value in field_data:
+            if value is None:
+                batch_data.append(None)
+            else:
+                # Skip isna check for iterables (except strings)
+                if not isinstance(value, (str, bytes)) and hasattr(value, '__iter__'):
+                    batch_data.append(value)
+                else:
+                    try:
+                        is_na = pd.isna(value)
+                        # Handle scalar values
+                        if isinstance(is_na, bool):
+                            batch_data.append(None if is_na else value)
+                        # Handle array-like values - just append as is
+                        else:
+                            batch_data.append(value)
+                    except:
+                        # If isna fails, just use the value as is
+                        batch_data.append(value)
+
+        return pc.cast(batch_data, arrow_type)
+
 
     def _construct_new_record_batch(self):
         self.current_record_batch.clear()
@@ -322,7 +325,7 @@ class BulkLoadWriter:
 
     def _create_next_file_writer(self):
         file_name = self._get_current_file_name()
-        if self.staging_config.type != 'oss':
+        if self.staging_config.type == 'cos':
             if not os.path.exists(os.path.dirname(file_name)):
                 self.file_io.create_dir(os.path.dirname(file_name))
         if self.file_format == FileFormatType.PARQUET:
@@ -365,6 +368,12 @@ class BulkLoadWriter:
                 raise e
         elif self.staging_config.type == 'oss':
             return self._get_current_file_name()
+        elif self.staging_config.type == 's3':
+            return self._get_current_file_name()
+        elif self.staging_config.type == 'gcs':
+            return self._get_current_file_name()
+        else:
+            return self._get_current_file_name()
 
     def _check_file_status(self):
         if self.writer is not None:
@@ -373,3 +382,96 @@ class BulkLoadWriter:
 
         if self.writer is None:
             self.writer = self._create_next_file_writer()
+
+
+def to_arrow_type(data_type: data_type_pb2.DataType) -> pa.DataType:
+    """Convert ClickZetta data type to Arrow type"""
+    category = data_type.category
+
+    if category == data_type_pb2.DataTypeCategory.INT8:
+        return pa.int8()
+    elif category == data_type_pb2.DataTypeCategory.INT16:
+        return pa.int16()
+    elif category == data_type_pb2.DataTypeCategory.INT32:
+        return pa.int32()
+    elif category == data_type_pb2.DataTypeCategory.INT64:
+        return pa.int64()
+    elif category == data_type_pb2.DataTypeCategory.FLOAT32:
+        return pa.float32()
+    elif category == data_type_pb2.DataTypeCategory.FLOAT64:
+        return pa.float64()
+    elif category == data_type_pb2.DataTypeCategory.DECIMAL:
+        if data_type.decimalTypeInfo:
+            precision = data_type.decimalTypeInfo.precision
+            scale = data_type.decimalTypeInfo.scale
+            return pa.decimal128(precision, scale)
+        return pa.decimal128(38, 10)  # Default precision and scale
+    elif category == data_type_pb2.DataTypeCategory.BOOLEAN:
+        return pa.bool_()
+    elif category in (data_type_pb2.DataTypeCategory.VARCHAR,
+                      data_type_pb2.DataTypeCategory.CHAR,
+                      data_type_pb2.DataTypeCategory.STRING):
+        return pa.string()
+    elif category == data_type_pb2.DataTypeCategory.BINARY:
+        return pa.binary()
+    elif category == data_type_pb2.DataTypeCategory.DATE:
+        return pa.date32()  # Days since Unix epoch
+    elif category == data_type_pb2.DataTypeCategory.TIMESTAMP_LTZ:
+        unit = _get_timestamp_unit(data_type.timestamp_info)
+        return pa.timestamp(unit, tz='UTC')
+    # TODO waiting for grpc proto update
+    # elif category == data_type_pb2.DataTypeCategory.TIMESTAMP_NTZ:
+    #     return pa.timestamp('us', tz=None)
+    elif category == data_type_pb2.DataTypeCategory.INTERVAL_DAY_TIME:
+        return pa.duration('us')
+    elif category == data_type_pb2.DataTypeCategory.INTERVAL_YEAR_MONTH:
+        return pa.duration('M')
+    elif category == data_type_pb2.DataTypeCategory.ARRAY:
+        element_type = to_arrow_type(data_type.arrayTypeInfo.elementType)
+        return pa.list_(element_type)
+    elif category == data_type_pb2.DataTypeCategory.MAP:
+        key_type = to_arrow_type(data_type.mapTypeInfo.keyType)
+        value_type = to_arrow_type(data_type.mapTypeInfo.valueType)
+        return pa.map_(key_type, value_type)
+    elif category == data_type_pb2.DataTypeCategory.STRUCT:
+        fields = []
+        for field in data_type.structTypeInfo.fields:
+            field_type = to_arrow_type(field.type)
+            fields.append(pa.field(field.name, field_type, field.type.nullable))
+        return pa.struct(fields)
+    # TODO waiting for grpc proto update
+    # elif category == data_type_pb2.DataTypeCategory.VECTOR_TYPE:
+    #     # Handle vector type as binary array
+    #     vector_info = data_type.vector_info
+    #     type_width = 0
+    #
+    #     if vector_info.numberType == data_type_pb2.VectorNumberType.I8:
+    #         type_width = 1  # Byte.BYTES
+    #     elif vector_info.numberType == data_type_pb2.VectorNumberType.I32:
+    #         type_width = 4  # Integer.BYTES
+    #     elif vector_info.numberType == data_type_pb2.VectorNumberType.F32:
+    #         type_width = 4  # Float.BYTES
+    #     else:
+    #         raise ValueError(f"Unsupported vector number type: {vector_info.numberType}")
+    #     # FixedSizeBinaryType
+    #     return pa.binary(vector_info.dimension * type_width)
+    #     # return pa.fixed_size_binary(vector_info.dimension * type_width)
+    else:
+        raise ValueError(f"Unsupported data type: {category}")
+
+
+def _get_timestamp_unit(timestamp_info: data_type_pb2.TimestampInfo) -> str:
+    """Get Arrow timestamp unit from proto timestamp info"""
+    if not timestamp_info:
+        return 'us'  # Default to microseconds
+
+    unit = timestamp_info.tsUnit
+    if unit == data_type_pb2.TimestampUnit.SECONDS:
+        return 's'
+    elif unit == data_type_pb2.TimestampUnit.MILLISECONDS:
+        return 'ms'
+    elif unit == data_type_pb2.TimestampUnit.MICROSECONDS:
+        return 'us'
+    elif unit == data_type_pb2.TimestampUnit.NANOSECONDS:
+        return 'ns'
+    return 'us'  # Default
