@@ -4,6 +4,7 @@ from __future__ import absolute_import
 from __future__ import division
 
 import re
+import sys
 import time
 import shutil
 from urllib.parse import urlparse, unquote
@@ -41,6 +42,8 @@ from clickzetta.query_result import QueryResult, QueryData, QueryDataType
 from clickzetta.proto.generated import ingestion_pb2
 from clickzetta.bulkload.bulkload_enums import BulkLoadOptions, BulkLoadMetaData, BulkLoadOperation, \
     BulkLoadCommitMode, BulkLoadConfig
+from clickzetta.utils import split_sql
+from clickzetta._volume import resolve_local_path
 
 TIMEOUT_HEADER = "X-Server-Timeout"
 TimeoutType = Union[float, None]
@@ -74,15 +77,15 @@ HEADERS = {
 
 DEFAULT_INSTANCE_ID = 100
 
-
 class Client(object):
 
     def __init__(self, username: str = None, password: str = None, instance: str = None, workspace: str = None,
                  vcluster: str = None, cz_url: str = None, service: str = None, schema: str = None,
-                 magic_token: str = None, protocol: str = None):
+                 magic_token: str = None, protocol: str = None, hints: {} = None):
         self.schema = None
         self.magic_token = None
         self.protocol = None
+        self.hints = hints
         if cz_url is not None:
             self._parse_url(cz_url)
             self.instance_id = 0
@@ -210,8 +213,9 @@ class Client(object):
             for key in options.record_keys:
                 keys.append(key)
             create_bulk_load_request.record_keys.extend(keys)
+        create_bulk_load_request.prefer_internal_endpoint = options.prefer_internal_endpoint
         response = self._gate_way_call(create_bulk_load_request, ingestion_pb2.MethodEnum.CREATE_BULK_LOAD_STREAM_V2)
-        response_pb = ParseDict(response, ingestion_pb2.CreateBulkLoadStreamResponse())
+        response_pb = ParseDict(response, ingestion_pb2.CreateBulkLoadStreamResponse(), ignore_unknown_fields=True)
         self.instance_id = response_pb.info.identifier.instance_id
         bulkload_meta_data = BulkLoadMetaData(response_pb.info.identifier.instance_id, response_pb.info)
 
@@ -244,7 +248,7 @@ class Client(object):
             commit_bulkload_request.commit_mode = ingestion_pb2.CommitBulkLoadStreamRequest.CommitMode.ABORT_STREAM
 
         response = self._gate_way_call(commit_bulkload_request, ingestion_pb2.MethodEnum.COMMIT_BULK_LOAD_STREAM_V2)
-        response_pb = ParseDict(response, ingestion_pb2.CommitBulkLoadStreamResponse())
+        response_pb = ParseDict(response, ingestion_pb2.CommitBulkLoadStreamResponse(), ignore_unknown_fields=True)
         bulkload_meta_data = BulkLoadMetaData(self.instance_id, response_pb.info)
         return bulkload_meta_data
 
@@ -267,13 +271,13 @@ class Client(object):
         get_bulkload_stream_request.stream_id = stream_id
         get_bulkload_stream_request.need_table_meta = True
         response = self._gate_way_call(get_bulkload_stream_request, ingestion_pb2.MethodEnum.GET_BULK_LOAD_STREAM_V2)
-        response_pb = ParseDict(response, ingestion_pb2.GetBulkLoadStreamResponse())
+        response_pb = ParseDict(response, ingestion_pb2.GetBulkLoadStreamResponse(), ignore_unknown_fields=True)
         bulkload_meta_data = BulkLoadMetaData(self.instance_id, response_pb.info)
         return bulkload_meta_data
 
     def open_bulkload_stream_writer(self, instance_id: int, workspace: string, schema_name: string, table_name: string,
                                     stream_id: string, partition_id: int) -> BulkLoadConfig:
-        open_bulkload_stream_requuest = ingestion_pb2.OpenBulkLoadStreamWriterRequest()
+        open_bulkload_stream_request = ingestion_pb2.OpenBulkLoadStreamWriterRequest()
         account = ingestion_pb2.Account()
         user_ident = ingestion_pb2.UserIdentifier()
         user_ident.instance_id = instance_id
@@ -281,18 +285,18 @@ class Client(object):
         user_ident.user_name = self.username
         account.user_ident.CopyFrom(user_ident)
         account.token = self.token
-        open_bulkload_stream_requuest.account.CopyFrom(account)
+        open_bulkload_stream_request.account.CopyFrom(account)
         table_identifier = ingestion_pb2.TableIdentifier()
         table_identifier.instance_id = instance_id
         table_identifier.workspace = workspace
         table_identifier.schema_name = schema_name
         table_identifier.table_name = table_name
-        open_bulkload_stream_requuest.identifier.CopyFrom(table_identifier)
-        open_bulkload_stream_requuest.stream_id = stream_id
-        open_bulkload_stream_requuest.partition_id = partition_id
-        response = self._gate_way_call(open_bulkload_stream_requuest,
+        open_bulkload_stream_request.identifier.CopyFrom(table_identifier)
+        open_bulkload_stream_request.stream_id = stream_id
+        open_bulkload_stream_request.partition_id = partition_id
+        response = self._gate_way_call(open_bulkload_stream_request,
                                        ingestion_pb2.MethodEnum.OPEN_BULK_LOAD_STREAM_WRITER_V2)
-        response_pb = ParseDict(response, ingestion_pb2.OpenBulkLoadStreamWriterResponse())
+        response_pb = ParseDict(response, ingestion_pb2.OpenBulkLoadStreamWriterResponse(), ignore_unknown_fields=True)
         bulkload_config = response_pb.config
         return BulkLoadConfig(bulkload_config)
 
@@ -321,7 +325,7 @@ class Client(object):
         finish_bulkload_stream_request.written_lengths.extend(written_lengths)
         response = self._gate_way_call(finish_bulkload_stream_request,
                                        ingestion_pb2.MethodEnum.FINISH_BULK_LOAD_STREAM_WRITER_V2)
-        response_pb = ParseDict(response, ingestion_pb2.FinishBulkLoadStreamWriterResponse())
+        response_pb = ParseDict(response, ingestion_pb2.FinishBulkLoadStreamWriterResponse(), ignore_unknown_fields=True)
         return response_pb.status
 
     def _gate_way_call(self, request, method: ingestion_pb2.MethodEnum):
@@ -334,6 +338,7 @@ class Client(object):
         HEADERS['X-ClickZetta-Token'] = self.token
         try:
             api_response = requests.post(self.service + path, data=MessageToJson(gate_way_request), headers=HEADERS)
+            api_response.encoding = 'utf-8'
             result = api_response.text
             result_dict = json.loads(result)
             if api_response.status_code != 200:
@@ -361,10 +366,35 @@ class Client(object):
             raise requests.exceptions.RequestException(
                 'gate_way_request error:{}'.format(e))
 
+    def pre_process_sql(self, sql):
+        queries = split_sql(sql)
+        if queries:
+            if len(queries) == 1 and self.process_use_cmd(queries[0]):
+                return None
+            for query in queries[:-1]:
+                self.process_use_cmd(query)
+        return queries[-1] + "\n;"
+
+    def process_use_cmd(self, sql: str):
+        has_processed = True
+        sql = re.sub(r'\s+', ' ', sql)
+        if sql.lower().startswith("use vcluster "):
+            self.vcluster = sql.lower().replace('use vcluster ', '').strip()
+        elif sql.lower().startswith('use schema '):
+            self.schema = sql.lower().replace('use schema ', '').strip()
+        elif sql.lower().startswith("use "):
+            self.schema = sql.lower().replace('use ', '').strip()
+        else:
+            has_processed = False
+        return has_processed
+
     def submit_sql_job(self, token: str, sql: str, job_id: JobID, parameters=None, schema=None) -> QueryResult:
+        sql = self.pre_process_sql(sql)
         path = "/lh/submitJob"
-        if 'test plain returns' in sql or 'test unicode returns' in sql:
-            return QueryResult({})
+        if (not sql) or 'test plain returns' in sql or 'test unicode returns' in sql:
+            res = QueryResult({})
+            res.data = QueryData(data=[], data_type=QueryDataType.Memory)
+            return res
         table = Table(self.workspace, '', self.instance, self.vcluster)
         vc = table.vcluster
         job_type = JobType.SQL_JOB
@@ -373,33 +403,43 @@ class Client(object):
         job_name = "SQL_JOB"
         user_id = 0
         reqeust_mode = JobRequestMode.HYBRID
-        timeout = FIRST_SQL_SUBMIT_TIMEOUT
+        polling_timeout = 30
         job_config = {}
-        sdk_job_timeout = 0
-        sdk_job_priority = 15
+        sdk_job_priority = 0
         sql_job_set_prop = {'cz.sql.adhoc.result.type': 'embedded', 'cz.sql.adhoc.default.format': 'arrow'}
+        sdk_job_timeout = 0
+        hints = self.hints if self.hints is not None else {}
         if parameters is not None and len(parameters) != 0:
-            if parameters['hints'] is not None:
-                for key in parameters['hints']:
-                    if key == 'sdk.job.timeout':
-                        sdk_job_timeout = int(parameters['hints'][key])
-                    elif key == 'sdk.job.priority':
-                        sdk_job_priority = int(parameters['hints'][key])
-                    else:
-                        sql_job_set_prop[key] = parameters['hints'][key]
-        timeout = sdk_job_timeout if 0 < sdk_job_timeout <= timeout else timeout
+            if 'hints' in parameters.keys():
+                hints = parameters['hints']
+            for key in parameters.keys():
+                if "%(" + key + ")s" in sql:
+                    sql = sql.replace("%(" + key + ")s", str(parameters[key]))
+        if len(hints) > 0:
+            for key in hints:
+                if key == 'sdk.job.timeout':
+                    sdk_job_timeout = int(hints[key])
+                elif key in ("sdk.job.priority", "schedule_job_queue_priority", "priority"):
+                    # The schedule_job_queue_priority and priority is adaptive to the java sdk set flag.
+                    sdk_job_priority = int(hints[key])
+                elif key in ('sdk.query.timeout.ms', 'querytimeout'):
+                    # The querytimeout is adaptive to the java sdk set flag.
+                    sdk_job_timeout = int(hints[key]) / 1000
+                else:
+                    sql_job_set_prop[key] = hints[key]
+        if 0 > polling_timeout or polling_timeout > FIRST_SQL_SUBMIT_TIMEOUT:
+            polling_timeout = FIRST_SQL_SUBMIT_TIMEOUT
         sql_config = SQLJobConfig(0, "0", "0", sql_job_set_prop)
         schema_name = ''
         if self.schema is not None:
             schema_name = self.schema
         if schema is not None:
             schema_name = schema
-        sql_job = SQLJob(sql_config, table.workspace, schema_name, [])
-        sql_job.query.append(sql)
-        job_timeout_ms = 3000
+        sql_job = SQLJob(sql_config, table.workspace, schema_name, [sql])
         user_agent = ""
-
-        job_desc = JobDesc(vc, job_type, job_id, job_name, user_id, reqeust_mode, timeout, job_config, sql_job,
+        max_value = sys.maxsize
+        job_timeout_ms = max_value if sdk_job_timeout >= max_value or sdk_job_timeout * 1000 >= max_value else sdk_job_timeout * 1000
+        job_desc = JobDesc(vc, job_type, job_id, job_name, user_id, reqeust_mode, polling_timeout, job_config, sql_job,
                            job_timeout_ms,
                            user_agent, sdk_job_priority)
         job_request = JobRequest(job_desc)
@@ -413,12 +453,14 @@ class Client(object):
                 raise requests.exceptions.RequestException(
                     "submit sql job failed:" + str(api_response.status_code) + api_response.text + ".sql:" + sql)
             api_response.encoding = 'utf-8'
-            result = api_response.text
-            result_dict = json.loads(result)
-            if int(result_dict['code']) != 200:
+            result = ""
+            try:
+                result = api_response.text
+                result_dict = json.loads(result)
+            except Exception as e:
                 raise requests.exceptions.RequestException(
-                    "submit sql job failed:" + str(result_dict))
-            get_job_result_dict = self.process_sql_response_with_timeout(sdk_job_timeout, result_dict, job_id)
+                    f"Parsing json data for job id [{job_id}] failed:{e}. raw result is: {result}")
+            get_job_result_dict = self.process_sql_response_with_timeout(result_dict, job_id, job_timeout_ms)
             logging.debug('GET RESPONSE FROM CZ SERVER FOR REQUEST:' + sql + ' TIME:' + str(datetime.now()))
             query_result = QueryResult(get_job_result_dict)
             if sql.lstrip().upper().startswith('PUT ') or sql.lstrip().upper().startswith('GET '):
@@ -434,11 +476,11 @@ class Client(object):
         if len(outcome) != 1:
             raise Exception(f"get volume sql failed, with result: {outcome}")
         outcome_obj = json.loads(outcome[0][0])
-        if outcome_obj['status'] is 'FAILED':
+        if outcome_obj['status'] == 'FAILED':
             raise Exception(f"{outcome_obj['request']['command']} volume failed: {outcome_obj['error']}")
         if upper_sql.startswith('GET'):
             query_result = self.gen_volume_result(outcome_obj)
-            if outcome_obj['status'] is 'CONTINUE' and outcome_obj['nextMarker'] is not '':
+            if outcome_obj['status'] == 'CONTINUE' and outcome_obj['nextMarker'] != '':
                 sql = 'set cz.sql.volume.file.transfer.next.marker=nextMarker;' + sql
                 job_id = JobID(self._format_job_id(), self.workspace, DEFAULT_INSTANCE_ID)
                 next_result = self.submit_sql_job(token=token, sql=sql, job_id=job_id)
@@ -502,7 +544,15 @@ class Client(object):
                 volume_files.append(path_parts[-1])
         return volume_files
 
-    def process_sql_response_with_timeout(self, sdk_job_timeout, result_dict, job_id):
+    def process_sql_response_with_timeout(self, result_dict, job_id, job_timeout_ms):
+        if ('status' in result_dict and 'errorCode' in result_dict['status'] and
+                result_dict['status']['errorCode'] == 'CZLH-60010'):
+            raise Exception("submit sql job:" + str(job_id.id) + " timeout after " + str(job_timeout_ms)
+                            + " ms. killed by lakehouse")
+        if job_timeout_ms == 0:
+            return self.wait_job_finished(result_dict, job_id, HEADERS)
+
+        sdk_job_timeout = job_timeout_ms / 1000.0
         if 0 < sdk_job_timeout <= FIRST_SQL_SUBMIT_TIMEOUT:
             # check resultSet first, then check status
             if 'resultSet' in result_dict:
@@ -519,11 +569,8 @@ class Client(object):
                         return result_dict
             else:
                 raise Exception("submit sql job:" + str(job_id.id) + " failed, root: " + str(result_dict))
-
-        elif sdk_job_timeout > FIRST_SQL_SUBMIT_TIMEOUT:
-            return self.wait_job_finished(result_dict, job_id, HEADERS, sdk_job_timeout - FIRST_SQL_SUBMIT_TIMEOUT)
         else:
-            return self.wait_job_finished(result_dict, job_id, HEADERS, HYBRID_SQL_TIMEOUT - FIRST_SQL_SUBMIT_TIMEOUT)
+            return self.wait_job_finished(result_dict, job_id, HEADERS, sdk_job_timeout)
 
     def close(self):
         pass
@@ -546,33 +593,68 @@ class Client(object):
         return False
 
     def get_job_profile(self, job_id: string):
+        result_dict = self._get_job(job_id, clickzetta.enums.JobRequestType.PROFILE)
+        return result_dict
+
+    def get_job_result(self, job_id: string):
+        result_dict = self._get_job(job_id, clickzetta.enums.JobRequestType.RESULT)
+        return result_dict
+
+    def get_job_progress(self, job_id: string):
+        result_dict = self._get_job(job_id, clickzetta.enums.JobRequestType.PROGRESS)
+        return result_dict
+
+    def get_job_summary(self, job_id: string):
+        result_dict = self._get_job(job_id, clickzetta.enums.JobRequestType.SUMMARY)
+        return result_dict
+
+    def get_job_plan(self, job_id: string):
+        result_dict = self._get_job(job_id, clickzetta.enums.JobRequestType.PLAN)
+        return result_dict
+
+    def _get_job(self, job_id: string, type: string):
         HEADERS['instanceName'] = self.instance
         HEADERS['X-ClickZetta-Token'] = self.token
         id = JobID(job_id, self.workspace, DEFAULT_INSTANCE_ID)
         account = clickzetta.enums.ClickZettaAccount(0)
-        get_job_profile_request = clickzetta.enums.GetJobProfileRequest(account, id, '')
-        api_request = clickzetta.enums.APIGetJobProfileRequest(get_job_profile_request, '')
+        get_job_request = clickzetta.enums.GetJobRequest(account, id, 0, '')
         path = '/lh/getJob'
-        data = json.dumps(api_request.to_api_repr())
         try:
+            api_request = clickzetta.enums.APIGetJobRequest(get_job_request, '', type)
+            data = json.dumps(api_request.to_api_repr())
             api_response = self.session.post(self.service + path, data=data, headers=HEADERS)
+            if not api_response.ok:
+                raise requests.exceptions.RequestException(f"Get job:{id.id} {type} failed:{api_response.text}")
             result = api_response.text
             result_dict = json.loads(result)
-            if int(result_dict['code']) != 200:
-                raise requests.exceptions.RequestException(
-                    "Get job :" + str(job_id.id) + " profile failed:" + result_dict)
             return result_dict
+        except AttributeError as e:
+            raise requests.exceptions.RequestException(f"Get job:{id.id} {type} failed:{e}, Only supports：result/progress/summary/plan/profile.")
         except Exception as e:
-            raise Exception("Get job:" + str(job_id.id) + " profile failed:" + str(e))
+            raise Exception(f"Get job:{id.id} {type} failed:{e}")
+
+    def is_job_finished(self, result_dict):
+        if 'status' in result_dict:
+            if 'state' in result_dict['status']:
+                status = result_dict['status']['state']
+                if status == "SUCCEED" or status == 'FAILED':
+                    return True
+                elif status == "QUEUEING" or status == "SETUP" or status == "RUNNING":
+                    if 'resultSet' in result_dict:
+                        if 'data' in result_dict['resultSet']:
+                            if 'data' in result_dict['resultSet']['data']:
+                                return True
+                        elif 'location' in result_dict['resultSet']:
+                            return True
+        return False
 
     def wait_job_finished(self, result_dict, job_id: JobID, headers, sdk_job_timeout=0):
         if 'status' in result_dict:
             if 'state' in result_dict['status']:
-                status = result_dict['status']['state']
-                if status == "QUEUEING" or status == "SETUP" or status == "RUNNING":
+                if not self.is_job_finished(result_dict):
                     account = clickzetta.enums.ClickZettaAccount(0)
-                    get_job_result_request = clickzetta.enums.GetJobResultRequest(account, job_id, 0, '')
-                    api_request = clickzetta.enums.APIGetJobRequest(get_job_result_request, '')
+                    get_job_result_request = clickzetta.enums.GetJobRequest(account, job_id, 0, '')
+                    api_request = clickzetta.enums.APIGetJobRequest(get_job_result_request, '', clickzetta.enums.JobRequestType.RESULT)
                     path = '/lh/getJob'
                     data = json.dumps(api_request.to_api_repr())
                     sql_job_start_time = datetime.now()
@@ -585,13 +667,11 @@ class Client(object):
                         get_job_result_response = self.session.post(self.service + path, data=data, headers=headers)
                         result = get_job_result_response.text
                         get_job_result_dict = json.loads(result)
-                        if int(get_job_result_dict['code']) != 200:
-                            raise requests.exceptions.RequestException(
-                                "Get async sql job:" + str(job_id.id) + " result failed:" + get_job_result_dict)
                         if 'status' in get_job_result_dict:
                             if 'state' in get_job_result_dict['status']:
                                 if get_job_result_dict['status']['state'] == 'SUCCEED' or \
-                                        get_job_result_dict['status']['state'] == 'FAILED':
+                                        get_job_result_dict['status']['state'] == 'FAILED' or \
+                                        get_job_result_dict['status']['state'] == 'CANCELLED':
                                     logging.info(
                                         "Get async sql job:" + str(job_id.id) + " result successfully.")
                                     return get_job_result_dict
@@ -608,7 +688,7 @@ class Client(object):
                                 "Get async sql job:" + str(
                                     job_id.id) + " result failed, http response job status is not exist.")
 
-                elif status == "SUCCEED" or status == 'FAILED':
+                else:
                     return result_dict
             else:
                 raise Exception("Execute sql job:" + str(job_id.id) + " failed, http response job state is not exist.")
@@ -621,6 +701,9 @@ class Client(object):
                            + str(random.randint(10000, 99999))
         return format_unique_id
 
+    def generate_job_id(self):
+        return self._format_job_id()
+
     def get_table_names(self, schema: str):
         job_id = JobID(self._format_job_id(), self.workspace, DEFAULT_INSTANCE_ID)
         query_result = self.submit_sql_job(token=self.token, sql='show tables;', schema=schema, job_id=job_id)
@@ -677,15 +760,15 @@ class Client(object):
             if option['name'].upper() == 'SOURCE_PREFIX':
                 prefix = option['value']
             option_part += ' ' + option['name'] + ' = ' + option['value']
-        local_file_list = self._get_real_local_file_list(outcome_obj['request']['localPaths'], prefix)
-        if len(local_file_list) == 0:
+        requested_paths = outcome_obj['request']['localPaths']
+        resolved_paths = []
+        for p in requested_paths:
+            resolved_paths += resolve_local_path(p)
+        if len(resolved_paths) == 0:
             raise Exception('No local file to put into volume')
-        sql = 'put '
-        for i in range(len(local_file_list)):
-            sql += "'" + local_file_list[i] + "'"
-            if i != len(local_file_list) - 1:
-                sql += ", "
-        sql += ' TO VOLUME ' + outcome_obj['request']['volume']
+        sql = 'PUT '
+        sql += ", ".join(f"'{p}'" for p in resolved_paths)
+        sql += ' TO ' + outcome_obj['request']['volumeIdentifier']
         if 'subdirectory' in outcome_obj['request']:
             sql += " SUBDIRECTORY '" + outcome_obj['request']['subdirectory'] + "'"
         elif 'file' in outcome_obj['request']:
@@ -694,19 +777,6 @@ class Client(object):
         sql += ';'
         return sql
 
-    def _get_real_local_file_list(self, local_path_list: [], prefix: str):
-        if len(local_path_list) == 0:
-            raise Exception('No local file to put into volume')
-        file_names = []
-        for local_path in local_path_list:
-            if local_path[-1] == '*' or local_path[-1] == '/':
-                local_dir = local_path if local_path[-1] == '/' else local_path[:-1]
-                file_names += [f for f in os.listdir(local_dir)
-                               if os.path.isfile(os.path.join(local_dir, f)) and f.startswith(prefix)]
-            else:
-                file_names.append(local_path)
-        return file_names
-
 
 def _add_server_timeout_header(headers: Optional[Dict[str, str]], kwargs):
     timeout = kwargs.get("timeout")
