@@ -1,12 +1,11 @@
 """Connection for ClickZetta DB-API."""
-
 import weakref
 import logging
 import time
 
 import requests
 from requests.adapters import HTTPAdapter
-from clickzetta import Client
+from clickzetta.client import Client
 from clickzetta.dbapi import cursor
 from clickzetta.dbapi import _helpers
 from clickzetta.bulkload.bulkload_enums import BulkLoadOperation, BulkLoadOptions, BulkLoadCommitOptions
@@ -85,7 +84,8 @@ class Connection(object):
 
     def cursor(self):
         """Return a new cursor object."""
-        self._update_token()
+        if self._client.username is not None and self._client.password is not None:
+            self._update_token()
         new_cursor = cursor.Cursor(self)
         self._cursors_created.add(new_cursor)
         return new_cursor
@@ -105,19 +105,33 @@ class Connection(object):
         vcluster = kwargs.get('vcluster', self._client.vcluster)
         partition_spec = kwargs.get('partition_spec')
         record_keys = kwargs.get('record_keys')
+        prefer_internal_endpoint = kwargs.get('prefer_internal_endpoint', False)
 
         bulkload_meta_data = self._client.create_bulkload_stream(
-            schema, table, BulkLoadOptions(operation, partition_spec, record_keys))
+            schema, table, BulkLoadOptions(operation, partition_spec, record_keys, prefer_internal_endpoint))
         return BulkLoadStream(bulkload_meta_data, self._client,
                               BulkLoadCommitOptions(workspace, vcluster))
 
     def get_bulkload_stream(self, stream_id: str, schema: str = None, table: str = None):
         bulkload_meta_data = self._client.get_bulkload_stream(schema, table, stream_id)
-        return BulkLoadStream(bulkload_meta_data, self._client)
+        return BulkLoadStream(bulkload_meta_data, self._client,
+                              BulkLoadCommitOptions(self._client.workspace, self._client.vcluster))
 
     def get_job_profile(self, job_id: str):
         return self._client.get_job_profile(job_id)
 
+    def get_job_result(self, job_id: str):
+        return self._client.get_job_result(job_id)
+
+    def get_job_progress(self, job_id: str):
+        return self._client.get_job_progress(job_id)
+
+    def get_job_summary(self, job_id: str):
+        return self._client.get_job_summary(job_id)
+
+    def get_job_plan(self, job_id: str):
+        return self._client.get_job_plan(job_id)
+
 
 def connect(**kwargs) -> Connection:
     client = kwargs.get('client')
@@ -127,5 +141,5 @@ def connect(**kwargs) -> Connection:
                         instance=kwargs.get('instance'), service=kwargs.get('service'),
                         workspace=kwargs.get('workspace'), vcluster=kwargs.get('vcluster'),
                         schema=kwargs.get('schema', 'public'), magic_token=kwargs.get('magic_token'),
-                        protocol=kwargs.get('protocol'))
+                        protocol=kwargs.get('protocol'), hints=kwargs.get('hints'))
     return Connection(client)
