@@ -92,6 +92,14 @@ class JobRequestMode(object):
     SYNC = "SYNC"
 
 
+class JobRequestType(object):
+    PLAN = "get_plan_request"
+    PROFILE = "get_profile_request"
+    RESULT = "get_result_request"
+    PROGRESS = "get_progress_request"
+    SUMMARY = "get_summary_request"
+
+
 class SQLJobConfig(object):
     def __init__(self, timeout, adhoc_size_limit: string, adhoc_row_limit: string, hint: map):
         self.timeout = timeout
@@ -137,7 +145,7 @@ class ClickZettaAccount(object):
         return copy.deepcopy(self._properties)
 
 
-class GetJobResultRequest(object):
+class GetJobRequest(object):
     def __init__(self, account: ClickZettaAccount, job_id: JobID, offset: int, user_agent: string):
         self.account = account
         self.job_id = job_id
@@ -185,10 +193,10 @@ class APIGetJobProfileRequest(object):
 
 
 class APIGetJobRequest(object):
-    def __init__(self, get_job_result_request: GetJobResultRequest, user_agent: string):
-        self.get_job_result_request = get_job_result_request
+    def __init__(self, get_job_request: GetJobRequest, user_agent: string, job_request_type: string):
+        self.get_job_result_request = get_job_request
         self.user_agent = user_agent
-        self._properties = {'get_result_request': get_job_result_request.to_api_repr(), 'user_agent': user_agent}
+        self._properties = {f'{job_request_type}': get_job_request.to_api_repr(), 'user_agent': user_agent}
 
     def to_api_repr(self) -> dict:
         return copy.deepcopy(self._properties)
@@ -216,6 +224,8 @@ class JobDesc(object):
                             'jobConfig': job_config, 'sqlJob': sql_job.to_api_repr()}
         if 'query_tag' in sql_job.sql_job_config.hint:
             self._properties['query_tag'] = sql_job.sql_job_config.hint['query_tag']
+        if job_timeout_ms > 0:
+            self._properties['jobTimeoutMs'] = job_timeout_ms
 
     def to_api_repr(self) -> dict:
         return copy.deepcopy(self._properties)
